from typing import List
from typing import Union

import pandas as pd

from tidybear.selectors import TidySelector
from tidybear.utils import get_column_names


def select(
    df: pd.DataFrame, *args: Union[str, TidySelector], **kwargs: str
) -> pd.DataFrame:
    """Select columns from a dataframe

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe to select columns from
    *args : str
        The column names to select
    **kwargs : str
        The column names to select and rename, new_name="old_name"

    Examples
    --------

    >>> df = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
    >>> select(df, "A")
       A
    0  1

    1  2
    2  3

    >>> select(df, "A", "A")
       A  A
    0  1  1
    1  2  2
    2  3  3

    >>> select(df, b="B")
       b
    0  4
    1  5
    2  6

    """

    to_select: List[Union[str, TidySelector]] = []

    if args:
        to_select.extend(args)

    if kwargs:
        to_select.extend(kwargs.values())
        rename_dict = {v: k for k, v in kwargs.items()}

    to_select_names = get_column_names(df.columns, to_select)
    selected = df.loc[:, to_select_names].copy()

    if kwargs:
        selected.rename(columns=rename_dict, inplace=True)

    return selected



from typing import Iterable
from typing import List
from typing import Sequence
from typing import Union

from tidybear.selectors import _ColumnList
from tidybear.selectors import TidySelector


def get_column_name(
    cols: Iterable[str], to_select: Union[str, TidySelector]
) -> Sequence[str]:
    if isinstance(to_select, str):
        return [to_select]

    return to_select(cols)


def get_column_names(cols: Iterable[str], to_select: _ColumnList) -> Sequence[str]:

    if isinstance(to_select, str) or isinstance(to_select, TidySelector):
        return get_column_name(cols, to_select)

    selected: List[str] = []
    for item in to_select:
        selected.extend(get_column_name(cols, item))

    return selected



from typing import Any

from pandas import DataFrame


def rename(df: DataFrame, *args: Any, **kwargs: Any) -> DataFrame:
    """Rename the columns of a dataframe

    You can use this function is a few different ways.
    - Use a list of strings to be used as the new column names.
      In this case the length of the list must equal the number
      of columns in the dataframe.
    - Use the new column names as arguments to the function.
      Again the number of arguments passed must equal the number
      of columns in the dataframe.
    - Use a dictionary with the keys as existing column names and
      values as the new column names.
    - Use keyword arguments with the key as existing column names
      and values as the new columns names.

    Parameters
    ----------
    df : DataFrame

    Returns
    -------
    DataFrame

    Examples
    --------

    ```
    >>> import pandas as pd
    >>> import tidybear as tb
    >>>
    >>> df = pd.DataFrame({"A": [1, 2],"B": [3, 4]})
    >>> df
    A  B
    0  1  3
    1  2  4
    >>> tb.rename(df, ["X", "Y"])
    X  Y
    0  1  3
    1  2  4
    >>> tb.rename(df, "X", "Y")
    X  Y
    0  1  3
    1  2  4
    >>> tb.rename(df, {"A": "X", "B": "Y"})
    X  Y
    0  1  3
    1  2  4
    >>> tb.rename(df, A="X", B="Y")
    X  Y
    0  1  3
    1  2  4
    ```
    """
    df = df.copy()
    if len(kwargs) > 0:
        return df.rename(columns=kwargs)

    if len(args) == 1 and isinstance(args[0], dict):
        return df.rename(columns=args[0])

    if len(args) >= 1:
        new_cols = args[0] if isinstance(args[0], list) else list(args)

        assert len(new_cols) == df.shape[1], (
            f"Number of columns provided ({len(new_cols)}) "
            f"does not match the number of features in the dataframe ({df.shape[1]})."
        )

        df.columns = new_cols
        return df

    return df



from tidybear.groupby import GroupBy
from tidybear.verbs.count import count
from tidybear.verbs.join import cross_join
from tidybear.verbs.join import inner_join
from tidybear.verbs.join import left_join
from tidybear.verbs.join import outer_join
from tidybear.verbs.join import right_join
from tidybear.verbs.mutate import mutate
from tidybear.verbs.pivot import pivot_longer
from tidybear.verbs.pivot import pivot_wider
from tidybear.verbs.rename import rename
from tidybear.verbs.select import select
from tidybear.verbs.slice import slice_max
from tidybear.verbs.slice import slice_min

__all__ = (
    "GroupBy",
    "count",
    "mutate",
    "pivot_longer",
    "pivot_wider",
    "rename",
    "slice_max",
    "slice_min",
    "select",
    "inner_join",
    "left_join",
    "right_join",
    "outer_join",
    "cross_join",
)



from __future__ import annotations

from typing import Any
from typing import List
from typing import Optional

import pandas as pd

from tidybear.selectors import _ColumnList
from tidybear.utils import get_column_names


class GroupBy:
    """Simplified API for performing groupby and summarise opperations in pandas.

    Parameters
    ----------
    df : DataFrame
        The dataframe to group.
    groups : str, List[str]
        Used to determine the groups for the groupby.

    Properties
    ----------
    groups : str, List[str]
        The grouping variables that were used.
    <df_columns> :
        You can access the columns of the DataFrameGroupBy object using the 'dot' syntax.
        To get the grouped columns with the string name, use `get` method.

    Functions
    ----------
    size : get group sizes
    get : get the grouped column by name
    summarise or summarize : cocatenate all active stats into a single dataframe.

    Examples
    ----------
    ```
    >>> import pandas as pd
    >>> import tidybear as tb
    >>>
    >>> df = pd.DataFrame({
    ...     "gr": list("AAABBBB"),
    ...     "val": [1, 2, 3, 7, 8, 8, 9]
    ... })
    >>>
    >>> with tb.GroupBy(df, "gr") as g:
    ...     tb.Stat.size()
    ...     summary = g.summarise()
        n
    gr
    A   3
    B   4
    ```
    """

    def __init__(self, df: pd.DataFrame, groups: _ColumnList) -> None:
        """Creates an active grouping that can track and summarise provided Stats.
        Must be used within a with statement.

        Parameters
        ----------
        df : DataFrame
            The dataframe to group.
        groups : str, List[str]
            Used to determine the groups for the groupby
        """
        self.__groups = get_column_names(df.columns, groups)
        self.__groupby_obj = df.groupby(self.__groups)

        self.__stats: List[pd.Series] = []

    def __enter__(self) -> GroupBy:
        return self

    def __exit__(self, *args: Any) -> None:
        self.__stats = []

    @property
    def groups(self) -> List[str]:
        """Get the grouping variables

        Returns
        -------
        str, List[str]
        """
        return list(self.__groups)

    def get(self, column: str) -> pd.Series:
        """Access a grouped column by name

        Parameters
        ----------
        column : str
            The name of the column

        Returns
        -------
        pd.Series
            The grouped column
        """
        return self.__groupby_obj[column]

    def summarise(self) -> pd.DataFrame:
        """Concatenate all active stats into a single dataframe.
        It will naturally join on the grouped columns.

        ```
        return pd.concat(active_stats, axis=1)
        ```

        Returns
        -------
        pd.DataFrame
            Final summary of all stats
        """
        return pd.concat(self.__stats, axis=1)

    def summarize(self) -> pd.DataFrame:
        """Concatenate all active stats into a single dataframe.
        It will naturally join on the grouped columns.

        ```
        return pd.concat(active_stats, axis=1)
        ```

        Returns
        -------
        pd.DataFrame
            Final summary of all stats
        """
        return self.summarise()

    def __add_stat(self, name: str, series: pd.Series) -> pd.Series:
        self.__stats.append(series.rename(name))
        return series.rename(name)

    def stat(self, name: str, series: pd.Series) -> pd.Series:
        return self.__add_stat(name, series)

    def n(self, name: Optional[str] = None) -> pd.Series:
        """Compute group sizes."""
        name = "n" if not name else name
        return self.__add_stat(name, self.__groupby_obj.size())

    def agg(
        self,
        column: str,
        func: Any,
        decimals: Optional[int] = None,
        name: Optional[str] = None,
        name_prefix: Optional[str] = None,
    ) -> pd.Series:
        """Aggregate one or more columns using one or more operations.

        Parameters
        ----------
        func : function, str, or list
            Function(s) to use for aggregating the data. See pd.Series.agg for acceptable strings.
        column : str or list
            Name of column to aggregate
        decimals : int, optional
            Number of decimals to round to, by default None
        name : str, optional
            New name of series, by default None. If none and the function provided is a string
            the name will be "{func}_{column}".
            If multiple funcions are provided this parameter is ignored.
        name_prefix: str, optional
            If passing multiple columns to agg with a custom function you can pass
            name_prefix to help name the summary columns
        temp : bool, optional, by default False
            If False, the Stat is appended to the GroupBy.
            If True, no Stat is appended, and instead the renamed series
            is returned for further operation.

        Returns
        -------
        Series or NoneType
        """

        if isinstance(func, list):
            for f in func:
                self.agg(column, f, decimals=decimals)
            return

        if isinstance(column, list):
            for c in column:
                self.agg(c, func, decimals=decimals)
            return

        if not name:
            if isinstance(func, str):
                name = func + "_" + column
            elif name_prefix:
                name = name_prefix + "_" + column
            else:
                name = column

        agg = self.get(column).agg(func)

        if decimals is not None:
            agg = agg.round(decimals)

        return self.__add_stat(name, agg)

    def n_distinct(self, column: str, **kwargs: Any) -> pd.Series:
        """Compute number of unique values in group."""
        if "name_prefix" not in kwargs:
            kwargs["name_prefix"] = "n_distinct"

        return self.agg(column, lambda x: len(x.unique()), **kwargs)

    def sum(self, column: str, **kwargs: Any) -> pd.Series:
        """Compute sum of group values."""
        return self.agg(column, "sum", **kwargs)

    def mean(self, column: str, **kwargs: Any) -> pd.Series:
        """Compute mean of group values."""
        return self.agg(column, "mean", **kwargs)

    def median(self, column: str, **kwargs: Any) -> pd.Series:
        """Compute median of group values."""
        return self.agg(column, "median", **kwargs)

    def max(self, column: str, **kwargs: Any) -> pd.Series:
        """Compute max of group values."""
        return self.agg(column, "max", **kwargs)

    def min(self, column: str, **kwargs: Any) -> pd.Series:
        """Compute min of group values."""
        return self.agg(column, "min", **kwargs)

    def var(self, column: str, **kwargs: Any) -> pd.Series:
        """Compute variance of group values."""
        return self.agg(column, "var", **kwargs)

    def std(self, column: str, **kwargs: Any) -> pd.Series:
        """Compute standard deviation of group values."""
        return self.agg(column, "std", **kwargs)

    def __str__(self) -> str:
        return f"GroupBy({self.groups})"



from typing import Any
from typing import List

import pandas as pd
from pandas import DataFrame


def join(
    left: pd.DataFrame,
    right: pd.DataFrame,
    how: str,
    *args: Any,
    **kwargs: str,
) -> pd.DataFrame:
    """Left join two dataframes on a column

    Parameters
    ----------
    left : pandas.DataFrame
        The left dataframe to join
    right : pandas.DataFrame
        The right dataframe to join
    *args : str
        The columns to join on
        Can be individual columns, or one list of columns
    **kwargs : str
        The columns to join, left="right"

    Returns
    -------
    pandas.DataFrame
        The joined dataframe

    """

    left_on: List[str] = []
    right_on: List[str] = []

    if args:
        if len(args) > 1 and any([isinstance(arg, list) for arg in args]):
            raise ValueError(
                "Only individual names or one list of names "
                "can be passed as unnamed args"
            )

        if len(args) == 1 and isinstance(args[0], list):
            left_on.extend(args[0])
            right_on.extend(args[0])
        else:
            left_on.extend(args)
            right_on.extend(args)

    if kwargs:
        left_on.extend(kwargs.keys())
        right_on.extend(kwargs.values())

    return left.merge(
        right,
        how=how,
        left_on=left_on,
        right_on=right_on,
    )


def inner_join(
    left: DataFrame, right: DataFrame, *args: Any, **kwargs: str
) -> DataFrame:
    """Left join two dataframes on a column

    Parameters
    ----------
    left : pandas.DataFrame
        The left dataframe to join
    right : pandas.DataFrame
        The right dataframe to join
    *args : str
        The columns to join on
        Can be individual columns, or one list of columns
    **kwargs : str
        The columns to join, left="right"

    Returns
    -------
    pandas.DataFrame
        The joined dataframe

    """
    return join(left, right, "inner", *args, **kwargs)


def left_join(
    left: DataFrame, right: DataFrame, *args: Any, **kwargs: str
) -> DataFrame:
    """Left join two dataframes on a column

    Parameters
    ----------
    left : pandas.DataFrame
        The left dataframe to join
    right : pandas.DataFrame
        The right dataframe to join
    *args : str
        The columns to join on
        Can be individual columns, or one list of columns
    **kwargs : str
        The columns to join, left="right"

    Returns
    -------
    pandas.DataFrame
        The joined dataframe

    """
    return join(left, right, "left", *args, **kwargs)


def right_join(
    left: DataFrame, right: DataFrame, *args: Any, **kwargs: str
) -> DataFrame:
    """Left join two dataframes on a column

    Parameters
    ----------
    left : pandas.DataFrame
        The left dataframe to join
    right : pandas.DataFrame
        The right dataframe to join
    *args : str
        The columns to join on
        Can be individual columns, or one list of columns
    **kwargs : str
        The columns to join, left="right"

    Returns
    -------
    pandas.DataFrame
        The joined dataframe

    """
    return join(left, right, "right", *args, **kwargs)


def outer_join(
    left: DataFrame, right: DataFrame, *args: Any, **kwargs: str
) -> DataFrame:
    """Left join two dataframes on a column

    Parameters
    ----------
    left : pandas.DataFrame
        The left dataframe to join
    right : pandas.DataFrame
        The right dataframe to join
    *args : str
        The columns to join on
        Can be individual columns, or one list of columns
    **kwargs : str
        The columns to join, left="right"

    Returns
    -------
    pandas.DataFrame
        The joined dataframe

    """
    return join(left, right, "outer", *args, **kwargs)


def cross_join(left: DataFrame, right: DataFrame) -> DataFrame:
    """Cross join two dataframes

    Parameters
    ----------
    left : pandas.DataFrame
        The left dataframe to join
    right : pandas.DataFrame
        The right dataframe to join
    """

    return left.merge(right, how="cross")



from __future__ import annotations

from typing import List
from typing import Union

from pandas import DataFrame


def _slice(
    df: DataFrame,
    order_by: str,
    n: int,
    ascending: bool,
    groupby: Union[str, List[str], None] = None,
) -> DataFrame:
    df = df.copy()

    if groupby:
        return (
            df.groupby(groupby)
            .apply(lambda x: _slice(x, order_by, n, ascending))
            .reset_index(drop=True)
        )

    return df.sort_values(order_by, ascending=ascending).head(n)


def slice_max(
    df: DataFrame,
    *,
    order_by: str,
    n: int,
    groupby: Union[str, List[str], None] = None,
) -> DataFrame:
    """Get the top N elements of a dataframe of group.

    Parameters
    ----------
    df : DataFrame
    order_by : str
        The column to order the values by
    n : int
        The number of elements to get
    groupby : str or list, optional
        Get top n elements by group. These columns used for groupby, by default None

    Returns
    -------
    Dataframe
    """
    return _slice(df, order_by, n, False, groupby)


def slice_min(
    df: DataFrame,
    *,
    order_by: str,
    n: int,
    groupby: Union[str, List[str], None] = None,
) -> DataFrame:
    """Get the bottom N elements of a dataframe of group.

    Parameters
    ----------
    df : DataFrame
    order_by : str
        The column to order the values by
    n : int
        The number of elements to get
    groupby : str or list, optional
        Get bottom n elements by group. These columns used for groupby, by default None

    Returns
    -------
    Dataframe
    """
    return _slice(df, order_by, n, True, groupby)



from typing import Any
from typing import Callable

from pandas import DataFrame


def mutate(df: DataFrame, **kwargs: Callable[..., Any]) -> DataFrame:
    """Create a new column in a dataframe using applied functions

    ```python
    tb.mutate(df, col_squared=lambda x: x.col**2)
    ```

    Parameters
    ----------
    df : DataFrame
    new_name : str
        the name of the new column to create
    definition : function
        the function to apply

    Returns
    -------
    DataFrame
    """

    df = df.copy()

    for name, definition in kwargs.items():
        df[name] = df.apply(definition, axis=1)

    return df



from __future__ import annotations

from pandas import DataFrame

from tidybear.selectors import _ColumnList
from tidybear.utils import get_column_names


def count(
    df: DataFrame,
    columns: _ColumnList,
    *,
    sort: bool = False,
    name: str = "n",
) -> DataFrame:
    """Quickly count the unique values of one or more variables.

    Parameters
    ----------
    df : DataFrame
        The dataframe to use
    columns : str, TidySelectors, or list or str, TidySelectors
        The column(s) to group by.
    sort : bool
        If True, will show the largest groups at the top, by default False
    name: str
        What to rename the new column with counts. By default "n" is used.
    """

    groupby_cols = get_column_names(df.columns, columns)
    counts = df.groupby(groupby_cols).size().rename(name).reset_index()

    if sort:
        return counts.sort_values(name, ascending=False)
    else:
        return counts.sort_values(columns)



from typing import Any
from typing import List
from typing import Optional
from typing import Union

import pandas as pd

from tidybear.selectors import _ColumnList
from tidybear.utils import get_column_names


def pivot_wider(
    df: pd.DataFrame,
    *,
    names_from: str = "name",
    values_from: Union[List[str], str] = "value",
    fill_value: Optional[Any] = None,
    prefix_names: bool = False,
) -> pd.DataFrame:
    """
    Transform a dataframe from long to wide

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe to transform
    names_from : str
        The column name to pivot on
    values_from : str, List[str]
        The column name to pivot on
        If multiple names are passed, the columns will be <value>_<name>
    fill_value : Optional[Any]
        The value to fill in the new column

    >>> df = pd.DataFrame({"idx": [1, 2], "name": ["a", "a"], "value": [3, 4]})
    >>> pivot_wider(df)
       idx  a
    0    1  3
    1    2  4

    Returns
    -------
    pandas.DataFrame
        The transformed dataframe
    """

    df = df.copy()

    if isinstance(values_from, str):
        values_from = [values_from]

    index_cols = [c for c in df.columns if c not in [names_from, *values_from]]
    df = df.pivot(index=index_cols, columns=names_from, values=values_from)

    if len(values_from) == 1 and not prefix_names:
        df.columns = [name for _, name in df.columns.to_flat_index()]
    else:
        df.columns = [f"{value}_{name}" for value, name in df.columns.to_flat_index()]

    if fill_value is not None:
        df = df.fillna(fill_value)

    return df.reset_index()


def pivot_longer(
    df: pd.DataFrame,
    cols: _ColumnList,
    *,
    names_to: str = "name",
    values_to: str = "value",
    drop_na: bool = True,
    cols_are_index: bool = False,
) -> pd.DataFrame:
    """
    Transform a dataframe from wide to long

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe to transform
    cols : str, list[str], TidySelector
        The columns to pivot on, use all the others as the index
    names_to : str
        The new name for the name column
    values_to : str
        the new name for the value column
    drop_na : bool, optional
        Whether to drop rows with missing values, default True
    cols_are_index : bool, optional
        Whether the columns are the index or the columns to pivot on, default False

    Examples
    --------

    >>> df = pd.DataFrame({"idx": [1, 2], "a": [1, 2], "b": [1, 2]})
    >>> pivot_longer(df, index="idx")
       idx name  value
    0    1    a      1
    1    1    b      1
    2    2    a      2
    3    2    b      2

    Returns
    -------
    pandas.DataFrame
        The transformed dataframe
    """

    df = df.copy()

    columns = get_column_names(df.columns, cols)
    index_columns = (
        columns if cols_are_index else [c for c in df.columns if c not in columns]
    )

    if len(index_columns) > 0:
        df.set_index(index_columns, inplace=True)

    df = df.stack(dropna=False).reset_index()
    df.columns = [*index_columns, names_to, values_to]

    if drop_na:
        df = df.dropna(subset=[values_to])

    return df



"""
Tidy Selectors

A tidy selector is a function that takes a list of column names,
and returns a new list of filtered column names based on the underlying tidy select function.

Examples
--------
code ::
    import pandas as pd
    import tidybear as tb
    from tidybear.selectors import contains

    data = pd.DataFrame(columns=["x01", "x02", "x03", "y1", "y2"])
    tb.select(data, contains("y"))
"""
from __future__ import annotations

import re
from typing import Iterable
from typing import List
from typing import Protocol
from typing import Sequence
from typing import Union


class TidySelectFunction(Protocol):
    def __call__(self, columns: Iterable[str]) -> Sequence[str]:
        ...


class TidySelector:
    def __init__(self, selector: TidySelectFunction) -> None:
        self.selector = selector

    def __call__(self, columns: Iterable[str]) -> Sequence[str]:
        return self.selector(columns)

    def filter_columns(self, columns: Iterable[str]) -> Sequence[str]:
        return self.selector(columns)

    def __neg__(self) -> TidySelector:
        def neg_selector(columns: Iterable[str]) -> List[str]:
            return [c for c in columns if c not in self.selector(columns)]

        return TidySelector(neg_selector)


_ColumnList = Union[str, TidySelector, Sequence[Union[str, TidySelector]]]


def everything() -> TidySelector:
    """Select all columns

    Examples
    --------
    >>> cols = ["x01", "x02", "x03", "y1", "y2"]
    >>> everything()(cols)
    ['x01', 'x02', 'x03', 'y1', 'y2']
    """

    def selector(columns: Iterable[str]) -> List[str]:
        return list(columns)

    return TidySelector(selector)


def last_col() -> TidySelector:
    """Select only the last column

    Examples
    --------
    >>> cols = ["x01", "x02", "x03", "y1", "y2"]
    >>> last_col()(cols)
    ['y2']
    """

    def selector(columns: Iterable[str]) -> List[str]:
        return [list(columns)[-1]]

    return TidySelector(selector)


def first_col() -> TidySelector:
    """Select only the first column

    Examples
    --------
    >>> cols = ["x01", "x02", "x03", "y1", "y2"]
    >>> first_col()(cols)
    ['x01']
    """

    def selector(columns: Iterable[str]) -> List[str]:
        return [list(columns)[0]]

    return TidySelector(selector)


def contains(pattern: str) -> TidySelector:
    """Select all columns that contain the literal string

    Parameters
    ----------
    pattern : str
        The string to match

    Examples
    --------
    >>> cols = ["x01", "x02", "x03", "y1", "y2"]
    >>> contains("y")(cols)
    ['y1', 'y2']
    """

    def selector(columns: Iterable[str]) -> List[str]:
        return [c for c in columns if pattern in c]

    return TidySelector(selector)


def matches(pattern: str) -> TidySelector:
    """Select all columns that match the regular expression

    pattern : str
        The regular expression to match

    Examples
    --------
    >>> cols = ["x01", "x02", "x03", "y1", "y2"]
    >>> matches("(x|y)0?2")(cols)
    ['x02', 'y2']
    """
    pattern_re = re.compile(pattern)

    def selector(columns: Iterable[str]) -> List[str]:
        return [c for c in columns if pattern_re.match(c)]

    return TidySelector(selector)


def starts_with(pattern: str) -> TidySelector:
    """Select all columns that start with the literal string

    Parameters
    ----------
    pattern : str
        The string to match

    Examples
    --------
    >>> cols = ["x01", "x02", "x03", "y1", "y2"]
    >>> starts_with("x")(cols)
    ['x01', 'x02', 'x03']
    """

    def selector(columns: Iterable[str]) -> List[str]:
        return [c for c in columns if c.startswith(pattern)]

    return TidySelector(selector)


def ends_with(pattern: str) -> TidySelector:
    """Select all columns that end with the literal srting

    Parameters
    ----------
    pattern : str
        The string to match

    Examples
    --------
    >>> cols = ["x01", "x02", "x03", "y1", "y2"]
    >>> ends_with("2")(cols)
    ['x02', 'y2']
    """

    def selector(columns: Iterable[str]) -> List[str]:
        return [c for c in columns if c.endswith(pattern)]

    return TidySelector(selector)


def num_range(prefix: str, values: Iterable[int], width: int = 0) -> TidySelector:
    """Select all columns that match a numeric range like x01, x02, x03

    Parameters
    ----------
    prefix : str
        The prefix of the columns to select

    values : range of integers
        The numeric values in the range

    width : int
        Pad the number with zeros to this width


    Examples
    --------
    >>> cols = ["x01", "x02", "x03", "y1", "y2"]
    >>> num_range("x", range(2, 4), width=2)(cols)
    ['x02', 'x03']
    """

    def selector(columns: Iterable[str]) -> List[str]:
        allowed = [f"{prefix}{str(i).zfill(width)}" for i in values]
        return [c for c in columns if c in allowed]

    return TidySelector(selector)
